@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
 from __future__ import absolute_import
 from __future__ import division
-import sys
 import json
 import math
 import fiona
@@ -12,26 +11,27 @@ from rasterio.transform import guard_transform
 from rasterio.enums import MaskFlags
 from affine import Affine
 import numpy as np
+from shapely import wkt, wkb
+
 try:
-    from shapely.errors import ReadingError
-except:
-    from shapely.geos import ReadingError
+    from shapely.errors import ShapelyError
+except ImportError:  # pragma: no cover
+    from shapely.errors import ReadingError as ShapelyError
+
 try:
     from json.decoder import JSONDecodeError
-except ImportError:
+except ImportError:  # pragma: no cover
     JSONDecodeError = ValueError
-from shapely import wkt, wkb
-from collections import Iterable, Mapping
+
+try:
+    from collections.abc import Iterable, Mapping
+except ImportError:  # pragma: no cover
+    from collections import Iterable, Mapping
 
 
 geom_types = ["Point", "LineString", "Polygon",
               "MultiPoint", "MultiLineString", "MultiPolygon"]
 
-PY3 = sys.version_info[0] >= 3
-if PY3:
-    string_types = str,  # pragma: no cover
-else:
-    string_types = basestring,  # pragma: no cover
 
 def wrap_geom(geom):
     """ Wraps a geometry dict in an GeoJSON Feature
@@ -58,14 +58,14 @@ def parse_feature(obj):
     try:
         shape = wkt.loads(obj)
         return wrap_geom(shape.__geo_interface__)
-    except (ReadingError, TypeError, AttributeError):
+    except (ShapelyError, TypeError, AttributeError):
         pass
 
     # wkb
     try:
         shape = wkb.loads(obj)
         return wrap_geom(shape.__geo_interface__)
-    except (ReadingError, TypeError):
+    except (ShapelyError, TypeError):
         pass
 
     # geojson-like python mapping
@@ -82,8 +82,7 @@ def parse_feature(obj):
 
 def read_features(obj, layer=0):
     features_iter = None
-    
-    if isinstance(obj, string_types):
+    if isinstance(obj, str):
         try:
             # test it as fiona data source
             with fiona.open(obj, 'r', layer=layer) as src:
@@ -159,6 +158,12 @@ def window_bounds(window, affine):
     return w, s, e, n
 
 
+def beyond_extent(window, shape):
+    """Checks if window references pixels beyond the raster extent"""
+    (wr_start, wr_stop), (wc_start, wc_stop) = window
+    return wr_start < 0 or wc_start < 0 or wr_stop > shape[0] or wc_stop > shape[1]
+
+
 def boundless_array(arr, window, nodata, masked=False):
     dim3 = False
     if len(arr.shape) == 3:
@@ -183,7 +188,7 @@ def boundless_array(arr, window, nodata, masked=False):
         window_shape = (wr_stop - wr_start, wc_stop - wc_start)
 
     # create an array of nodata values
-    out = np.empty(shape=window_shape)
+    out = np.empty(shape=window_shape, dtype=arr.dtype)
     out[:] = nodata
 
     # Fill with data where overlapping
@@ -263,8 +268,8 @@ class Raster(object):
         col, row = [math.floor(a) for a in (~self.affine * (x, y))]
         return row, col
 
-    def read(self, bounds=None, window=None, masked=False):
-        """ Performs a boundless read against the underlying array source
+    def read(self, bounds=None, window=None, masked=False, boundless=True):
+        """ Performs a read against the underlying array source
 
         Parameters
         ----------
@@ -276,6 +281,9 @@ class Raster(object):
         masked: boolean
             return a masked numpy array, default: False
             bounds OR window are required, specifying both or neither will raise exception
+        boundless: boolean
+            allow window/bounds that extend beyond the dataset’s extent, default: True
+            partially or completely filled arrays will be returned as appropriate.
 
         Returns
         -------
@@ -292,6 +300,9 @@ class Raster(object):
         else:
             raise ValueError("Specify either bounds or window")
 
+        if not boundless and beyond_extent(win, self.shape):
+            raise ValueError("Window/bounds is outside dataset extent and boundless reads are disabled")
+
         c, _, _, f = window_bounds(win, self.affine)  # c ~ west, f ~ north
         a, b, _, d, e, _, _, _, _ = tuple(self.affine)
         new_affine = Affine(a, b, c, d, e, f)
@@ -313,7 +324,7 @@ class Raster(object):
                     warnings.warn("Setting masked to True because dataset mask has been detected")
 
             new_array = self.src.read(
-                self.band, window=win, boundless=True, masked=masked)
+                self.band, window=win, boundless=boundless, masked=masked)
 
         return Raster(new_array, new_affine, nodata)
 
